/*---------------------------------- Title -----------------------------------\
|  KA1 File extractor for K5000 bank (KAA) files.                             |
|  Programmer: Jens Groh                                                      |
|  Version: 0.3 beta                                                          |
|  Date: 11.07.1999                                                           |
\----------------------------------------------------------------------------*/

/*------------------------- Preprocessor Directives -------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*--------------------------------- Switches --------------------------------*/

/*----------------------------- Global Constants ----------------------------*/
#define POOL_SIZE             0x20000
#define PATCHES               128
#define SOURCES               6

#define TONE_COMMON_DATA_SIZE 82
#define NAME_OFFSET           40
#define NAME_SIZE             8
#define SRC_NMBR_OFFSET       51

#define SOURCE_DATA_SIZE      86
#define WAVE_KIT_MSB_OFFSET   28
#define WAVE_KIT_ADD_MASK     0x04

#define ADD_WAVE_KIT_SIZE     806

#ifndef FALSE
#define FALSE                 0
#endif

#ifndef TRUE
#define TRUE                  (!0)
#endif

/*----------------------------- Type Definitions ----------------------------*/
typedef struct {
   int is_add;
   unsigned int add_kit_ptr;
} source_type;

typedef struct {
   int index;
   int is_used;
   unsigned int tone_ptr;
   int sources;
   int add_kits;
   source_type source[SOURCES];
   int size;
   char name[NAME_SIZE+1];
   char code[SOURCES+1];
} patch_type;

typedef struct {
   unsigned char data_pool[POOL_SIZE];
   patch_type patch[PATCHES+1];
   int patches;
   unsigned int base;
   patch_type *used_patch[PATCHES+1];
} bank_type;

/*----------------------------- Global Variables ----------------------------*/
static char *exitstr = "";

/*---------------------------- Function Prototypes --------------------------*/
unsigned int get_ptr(FILE *fp);
void get_bank(bank_type *bankp, FILE *fp);
int compare_tone_ptr(const patch_type **p1, const patch_type **p2);
void exit_func0(void);

/*----------------------------- Inline Functions ----------------------------*/

/*------------------------------- Main Program ------------------------------*/
int main(int argc, char **argv) {
   int i;
   char *infilename = NULL;
   char *outpathname = NULL;
   char ka1filename[513] = "";
   char patchname[NAME_SIZE+1] = "";
   int do_list = FALSE;
   int wrongparam = FALSE;
   FILE *infile = NULL;
   FILE *ka1file = NULL;
   bank_type bank;
   int p;
   patch_type *patchp;

   atexit(exit_func0);
   
   /* parse command line parameters: */
   for (i = 1; i < argc; i++) {
      if (*argv[i] == '-') {
         switch (*(argv[i] + 1)) {
         case 'i':
         case 'I':
            infilename = argv[i] + 2;
            break;
         case 'e':
         case 'E':
            outpathname = argv[i] + 2;
            break;
         case 'l':
         case 'L':
            do_list = TRUE;
            break;
         default:
            wrongparam = TRUE;
         }
      } 
   }
   if (infilename == NULL) {
      wrongparam = TRUE;
   }
   /* display help: */
   if (wrongparam) {
      exitstr = "\
kaatoka1 usage:\n\
\n\
kaatoka1 -i<infile> [-e<outpath>] [-l]\n\
Parameters: (in any order, lower or upper case)\n\
  -i<infile>     : read <K5000 bank (.KAA) file>\n\
  -e<outpath>    : extract .KA1 files to <destination path>\n\
  -l             : list patches\n\
";
      exit(EXIT_FAILURE);
   }

   /* read and analyze KAA file: */
   if ((infile = fopen(infilename,"rb")) == NULL) {
      exitstr = "Cannot open input file.";
      exit(EXIT_FAILURE);
   }
   get_bank(&bank, infile);
   fclose(infile);
   printf("\"%s\" contains %d patches. %d bytes (%5.2f%% of memory) free.\n", infilename, bank.patches, POOL_SIZE - bank.used_patch[bank.patches]->tone_ptr, (float)(POOL_SIZE - bank.used_patch[bank.patches]->tone_ptr) * 100. / (float)POOL_SIZE);

   /* list the patches: */
   if (do_list) {
      printf("number name     sources  size\n");
      for (p = 0; p < bank.patches; p++) {
         patchp = bank.used_patch[p];
         printf(" % 4d  %8s  % 6s % 5d\n", patchp->index + 1, patchp->name, patchp->code, patchp->size);
      }
   }

   /* extract KA1 files: */
   if (outpathname != NULL) {
      if (outpathname[strlen(outpathname) - 1] != '\\') {
         strcat(outpathname,"\\");
      }
      printf("\nExtracting to path \"%s\"...",outpathname);
      for (p = 0; p < bank.patches; p++) {
         patchp = bank.used_patch[p];
         /* remove trailing spaces from name: */
         strcpy(patchname,patchp->name);
         for (i = NAME_SIZE - 1; i >= 0 ; i--) {
            if (patchname[i] != ' ') {
               break;
            }
            patchname[i] = '\0';
         }
         printf("\n%s.KA1",patchname);
         strcat(strcat(strcpy(ka1filename,outpathname),patchname),".KA1");
         if ((ka1file = fopen(ka1filename,"wb+")) == NULL) {
            printf(" - Cannot open KA1 file! (Patch name may be incompatible.)");
         } else {
            fwrite(&bank.data_pool[patchp->tone_ptr], 1, patchp->size, ka1file);
            fclose(ka1file);
            ka1file = NULL;
         }
      }
      printf("\n");
   }

   exit(EXIT_SUCCESS);
   return 0;
}

/*--------------------------------- Function ---------------------------------\
| Read a single pointer table entry from a KAA file.                          |
\----------------------------------------------------------------------------*/
unsigned int get_ptr(FILE *fp) {
   unsigned int ptr;

   /* K5000 pointers are MSByte first */
   ptr = (unsigned int)(unsigned char)fgetc(fp) << 24;
   ptr |= (unsigned int)(unsigned char)fgetc(fp) << 16;
   ptr |= (unsigned int)(unsigned char)fgetc(fp) << 8;
   ptr |= (unsigned int)(unsigned char)fgetc(fp);
   return ptr;
}

/*--------------------------------- Function ---------------------------------\
| Read a KAA file to a bank struct, adjust pointers and perform an analysis.  |
\----------------------------------------------------------------------------*/
void get_bank(bank_type *bankp, FILE *fp) {
   int p, s;
   patch_type *patchp;
   source_type *sourcep;
   int patches;
   int i;
   char *c, *pp;

   /* read the pointer table (128 * 7 pointers) and build list of used patches: */
   patches = 0;
   for (p = 0, patchp = bankp->patch; p < PATCHES; p++, patchp++) {
      /* 1 pointer to tone data */
      patchp->index = p;
      patchp->tone_ptr = get_ptr(fp);
      patchp->is_used = (patchp->tone_ptr != 0);
      if (patchp->is_used) {
         bankp->used_patch[patches] = patchp;
         patches++;
      }
      /* 6 pointers to ADD wave kits */
      patchp->add_kits = 0;
      for (s = 0, sourcep = patchp->source; s < SOURCES; s++, sourcep++) {
         sourcep->add_kit_ptr = get_ptr(fp);
         sourcep->is_add = (sourcep->add_kit_ptr != 0);
         if (sourcep->is_add) {
            patchp->add_kits++;
         }
      }
   }
   /* read the 'memory high water mark' pointer: */
   patchp->index = p;
   patchp->tone_ptr = get_ptr(fp);
   bankp->used_patch[patches] = patchp;
   bankp->patches = patches;
   /* read the data pool: */
   fread(bankp->data_pool, 1, POOL_SIZE, fp);
   /* adjust non-zero pointers: */
   qsort(bankp->used_patch, bankp->patches, sizeof(patch_type *), compare_tone_ptr);
   bankp->base = bankp->used_patch[0]->tone_ptr;
   for (p = 0; p < bankp->patches; p++) {
      patchp = bankp->used_patch[p];
      patchp->tone_ptr -= bankp->base;
      for (s = 0, sourcep = patchp->source; s < SOURCES; s++, sourcep++) {
         if (sourcep->is_add) {
            sourcep->add_kit_ptr -= bankp->base;
         }
      }
   }
   bankp->patch[PATCHES].tone_ptr -= bankp->base;
   /* analyze number of sources, size and name: */
   for (p = 0; p < bankp->patches; p++) {
      patchp = bankp->used_patch[p];
      patchp->sources = bankp->data_pool[patchp->tone_ptr + SRC_NMBR_OFFSET];
      patchp->size = TONE_COMMON_DATA_SIZE + SOURCE_DATA_SIZE * patchp->sources + ADD_WAVE_KIT_SIZE * patchp->add_kits;
      c = patchp->name;
      pp = &bankp->data_pool[patchp->tone_ptr + NAME_OFFSET];
      for (i = 0; i < NAME_SIZE; i++) {
         *c++ = *pp++;
      }
      *c = '\0';
   }
   /* analyze tone structure: */
   for (p = 0; p < bankp->patches; p++) {
      patchp = bankp->used_patch[p];
      for (s = 0, sourcep = patchp->source; s < patchp->sources; s++, sourcep++) {
         patchp->code[s] = sourcep->is_add ? 'A' : 'P';
      }
      while (s < SOURCES) {
         patchp->code[s++] = '-';
      }
      patchp->code[s] = '\0';
   }
}

/*--------------------------------- Function --------------------------------*/
int compare_tone_ptr(const patch_type **p1, const patch_type **p2) {
   return (*p1)->tone_ptr - (*p2)->tone_ptr;
}

/*--------------------------------- Function --------------------------------*/
void exit_func0(void) {
   fcloseall();
   printf("\n%s\n",exitstr);
}

/*---------------------------------------------------------------------------*/
